import math

EPS = 1e-8


class Value:

    def __init__(self, data, _children=(), _op="", label=""):
        self.data = data
        self.grad = 0.0
        self._backward = lambda: None
        self._prev = set(_children)
        self._op = _op
        self.label = label

    def _clip(self, data):
        return max(EPS, min(1 - EPS, self.data))

    def __repr__(self):
        return f"Value(data={self.data})"

    def __add__(self, other):
        other = other if isinstance(other, Value) else Value(other)
        out = Value(self.data + other.data, (self, other), "+")

        def _backward():
            self.grad += 1.0 * out.grad
            other.grad += 1.0 * out.grad

        out._backward = _backward

        return out

    def __mul__(self, other):
        other = other if isinstance(other, Value) else Value(other)
        out = Value(self.data * other.data, (self, other), "*")

        def _backward():
            self.grad += other.data * out.grad
            other.grad += self.data * out.grad

        out._backward = _backward

        return out

    def __pow__(self, other):
        other = other if isinstance(other, Value) else Value(other)
        out = Value(self.data**other.data, (self, other), f"**")

        def _backward():
            self.grad += other.data * (self.data ** (other.data - 1)) * out.grad
            if self.data > 0:  # log is only valid for positive bases
                other.grad += (self.data**other.data) * math.log(self.data) * out.grad

        out._backward = _backward

        return out

    def __rmul__(self, other):  # other * self
        return self * other

    def __truediv__(self, other):  # self / other
        return self * other**-1

    def __neg__(self):  # -self
        return self * -1

    def __sub__(self, other):  # self - other
        return self + (-other)

    def __rsub__(self, other):  # other - self
        return other + (-self)

    def __radd__(self, other):  # other + self
        return self + other

    def tanh(self):
        x = self.data
        t = (math.exp(2 * x) - 1) / (math.exp(2 * x) + 1)
        out = Value(t, (self,), "tanh")

        def _backward():
            self.grad += (1 - t**2) * out.grad

        out._backward = _backward

        return out

    def sigmoid(self):
        x = self.data
        s = 1 / (1 + math.exp(-x))
        out = Value(s, (self,), "sigmoid")

        def _backward():
            self.grad += s * (1 - s) * out.grad

        out._backward = _backward

        return out

    def relu(self):
        x = self.data
        r = max(0, x)
        out = Value(r, (self,), "relu")

        def _backward():
            self.grad += (x > 0) * out.grad

        out._backward = _backward

        return out

    def leaky_relu(self, negative_slope=0.01):
        x = self.data
        r = max(0, x) + negative_slope * min(0, x)
        out = Value(r, (self,), "leaky_relu")

        def _backward():
            self.grad += (x > 0) * out.grad
            self.grad += (x <= 0) * negative_slope * out.grad

        out._backward = _backward

        return out

    def log(self):
        x = self.data
        out = Value(math.log(self._clip(x)), (self,), "log")

        def _backward():
            self.grad += (1 / self._clip(x)) * out.grad

        out._backward = _backward

        return out

    def exp(self):
        x = self.data
        out = Value(math.exp(x), (self,), "exp")

        def _backward():
            self.grad += out.data * out.grad

        out._backward = _backward

        return out

    def backward(self):

        topo = []
        visited = set()

        def build_topo(v):
            if v not in visited:
                visited.add(v)
                for child in v._prev:
                    build_topo(child)
                topo.append(v)

        build_topo(self)

        self.grad = 1.0
        for node in reversed(topo):
            node._backward()



from mikegrad.engine import Value
from typing import List
import math
import numpy as np


def from_pandas(df):
    """
    Generate a list of Value objects from a pandas DataFrame.
    """
    return from_numpy(df.values)


def from_numpy(array):
    """
    Generate a list of Value objects from a numpy array.
    """
    return from_list(array.tolist())


def to_numpy(values):
    """
    Generate a numpy array from a list of Value objects.
    """
    return np.array(to_list(values))


def from_list(values):
    """
    Generate a list of Value objects from a N dimension list of numbers.
    """
    if type(values) == list:
        return [from_list(value) for value in values]
    return Value(values)


def to_list(values):
    """
    Generate a N dimension list of numbers from a list of Value objects.
    """
    if type(values) == list:
        return [to_list(value) for value in values]
    return values.data


def one_hot_encode(values: List[Value], num_classes: int):
    one_hot = []
    for value in values:
        one_hot.append([Value(0) for _ in range(num_classes)])
        one_hot[-1][int(value.data)] = Value(1)
    return one_hot


def argmax(values):
    max_value = -math.inf
    max_index = -1
    for i, value in enumerate(values):
        if value.data > max_value:
            max_value = value.data
            max_index = i

    return max_index


def log(value: Value):
    return value.log()


def exp(value: Value):
    return value.exp()


def sigmoid(value: Value):
    return value.sigmoid()


def tanh(value: Value):
    return value.tanh()


def relu(value: Value):
    return value.relu()


def leaky_relu(value: Value, negative_slope=0.01):
    return value.leaky_relu(negative_slope)


def train(mlp, xs, ys, loss_fn, optimizer, nepochs=100, batch_size=-1, use_tqdm=True):
    if batch_size == -1:
        batch_size = len(xs)

    if use_tqdm:
        from tqdm import tqdm

        epoch_progress = tqdm(range(nepochs), desc="Epochs")

    for i in range(nepochs):
        epoch_loss = 0.0
        n = 0

        use_batch_progress = use_tqdm and batch_size != len(xs)
        if use_batch_progress:
            len_batches = math.ceil(len(xs) / batch_size)
            batch_progress = tqdm(range(len_batches), desc="Batches", leave=False)

        for j in range(0, len(xs), batch_size):
            x_batch = xs[j : j + batch_size]
            y_batch = ys[j : j + batch_size]
            optimizer.zero_grad()

            y_pred = [mlp(x) for x in x_batch]
            loss = loss_fn(y_pred, y_batch)
            epoch_loss += loss.data
            n += 1

            loss.backward()
            optimizer.step()
            if use_batch_progress:
                batch_progress.set_postfix(loss=f"{(loss.data):.6f}")
                batch_progress.update(1)

        if use_tqdm:
            epoch_progress.set_postfix(loss=f"{(epoch_loss/n):.6f}")
            epoch_progress.update(1)
        else:
            print(f"Epoch [{i+1}/{nepochs}] | Loss: {(epoch_loss/n):.6f}")


def test(mlp, xs, ys, loss_fn, use_tqdm=True):
    total_loss = 0.0
    n = 0
    y_preds = []

    if use_tqdm:
        from tqdm import tqdm

        progress = tqdm(range(len(xs)), desc="Testing")

    for i in range(len(xs)):
        x = xs[i]
        y = ys[i]

        y_pred = mlp(x)
        y_preds.append(y_pred)
        loss = loss_fn([y_pred], [y])
        total_loss += loss.data
        n += 1

        if use_tqdm:
            progress.set_postfix(loss=f"{(loss.data):.6f}")
            progress.update(1)

    return total_loss / n, y_preds



import random
from mikegrad.engine import Value
import mikegrad as mg


class BaseModule:
    def __call__(self, *args):
        return self.forward(*args)


class Identity(BaseModule):
    def forward(self, x):
        return x


class Sigmoid(BaseModule):
    def forward(self, x):
        if type(x) == list:
            return [mg.sigmoid(xi) for xi in x]
        return mg.sigmoid(x)


class Tanh(BaseModule):
    def forward(self, x):
        if type(x) == list:
            return [mg.tanh(xi) for xi in x]
        return mg.tanh(x)


class ReLU(BaseModule):
    def forward(self, x):
        if type(x) == list:
            return [mg.relu(xi) for xi in x]
        return mg.relu(x)


class LeakyReLU(BaseModule):
    def __init__(self, negative_slope=0.01):
        super().__init__()
        self.negative_slope = negative_slope

    def forward(self, x):
        if type(x) == list:
            return [mg.leaky_relu(xi, self.negative_slope) for xi in x]
        return mg.leaky_relu(x, self.negative_slope)


class Softmax(BaseModule):
    def forward(self, x):
        exps = [mg.exp(xi) for xi in x]
        exp_sum = sum(exps)
        return [exp_i / exp_sum for exp_i in exps]


class Neuron(BaseModule):

    def __init__(self, nin, act=None):  # nin: number of inputs
        super().__init__()
        self.w = [Value(random.uniform(-1, 1)) for _ in range(nin)]
        self.b = Value(random.uniform(-1, 1))
        self.act = act
        if act is None:
            self.act = Identity()

    def forward(self, x):
        # w * x + b
        out = sum([wi * xi for wi, xi in zip(self.w, x)], self.b)
        out = self.act(out)
        return out

    def parameters(self):
        return self.w + [self.b]


class Layer(BaseModule):

    def __init__(self, nin, nout, act=None, layer_act=None):
        super().__init__()
        if act is None:
            act = Identity()
        self.layer_act = layer_act
        self.neurons = [Neuron(nin, act) for _ in range(nout)]

    def forward(self, x):
        outs = [n(x) for n in self.neurons]
        if self.layer_act is not None:
            outs = self.layer_act(outs)
        return outs[0] if len(outs) == 1 else outs

    def parameters(self):
        return [p for n in self.neurons for p in n.parameters()]


class MLP(BaseModule):
    def __init__(self, nin, nouts, act=None, out_act=None):
        super().__init__()
        if act is None:
            act = Sigmoid()
        if out_act is None:
            out_act = Identity()

        sizes = [nin] + nouts  # [2] + [3, 4] = [2, 3, 4]
        self.layers = [
            Layer(
                sizes[i],
                sizes[i + 1],
                act if i != len(nouts) - 1 else None,
                out_act if i == len(nouts) - 1 else None,
            )
            for i in range(len(nouts))
        ]  # len(nouts) = len(sizes) - 1

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self):
        return [p for layer in self.layers for p in layer.parameters()]



from typing import List
from mikegrad.engine import Value


class BaseOptimizer:
    def __init__(self):
        pass

    def step(self):
        raise NotImplementedError

    def zero_grad(self):
        for param in self.params:
            if param.grad is not None:
                param.grad = 0


class SGD(BaseOptimizer):
    def __init__(
        self,
        params: List[Value],
        lr: float = 0.001,
        momentum: float = 0,
        dampening: float = 0,
        weight_decay: float = 0,
        nesterov: bool = False,
        maximize: bool = False,
    ):
        super().__init__()
        self.params = params
        self.lr = lr
        self.momentum = momentum
        self.dampening = dampening
        self.weight_decay = weight_decay
        self.nesterov = nesterov
        self.maximize = maximize
        self.momentum_buffers = [None for _ in params]  # One buffer per parameter

    def step(self):
        for i, param in enumerate(self.params):
            if param.grad is None:
                continue

            grad = param.grad

            if self.weight_decay != 0:
                grad = grad + self.weight_decay * param.data

            if self.momentum != 0:
                if self.momentum_buffers[i] is None:
                    self.momentum_buffers[i] = grad
                else:
                    buf = self.momentum_buffers[i]
                    buf = self.momentum * buf + (1 - self.dampening) * grad
                    self.momentum_buffers[i] = buf

                if self.nesterov:
                    grad = grad + self.momentum * buf
                else:
                    grad = buf

            update = self.lr * grad
            if self.maximize:
                param.data += update
            else:
                param.data -= update

        return self.params


class Adam(BaseOptimizer):
    def __init__(
        self,
        params,
        lr=0.001,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0,
        amsgrad=False,
        maximize=False,
    ):
        super().__init__()
        self.params = params
        self.lr = lr
        self.B1, self.B2 = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.amsgrad = amsgrad
        self.maximize = maximize

        self.m = [0 for _ in params]
        self.v = [0 for _ in params]
        self.vmax = [0 for _ in params]

        self.t = 0

    def step(self):
        self.t += 1
        for i, param in enumerate(self.params):
            if param.grad is None:
                continue

            grad = param.grad

            if self.maximize:
                grad = -grad

            if self.weight_decay != 0:
                grad = grad + self.weight_decay * param.data

            self.m[i] = self.B1 * self.m[i] + (1 - self.B1) * grad
            self.v[i] = self.B2 * self.v[i] + (1 - self.B2) * (grad**2)

            m_hat = self.m[i] / (1 - (self.B1**self.t))
            v_hat = self.v[i] / (1 - (self.B2**self.t))

            if self.amsgrad:
                self.vmax[i] = max(self.vmax[i], v_hat)
                param.data -= self.lr * m_hat / (self.vmax[i] ** 0.5 + self.eps)
            else:
                param.data -= self.lr * m_hat / (v_hat**0.5 + self.eps)


class AdamW(BaseOptimizer):
    def __init__(
        self,
        params,
        lr=0.001,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0.01,
        amsgrad=False,
        maximize=False,
    ):
        super().__init__()
        self.params = params
        self.lr = lr
        self.B1, self.B2 = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.amsgrad = amsgrad
        self.maximize = maximize

        self.m = [0 for _ in params]
        self.v = [0 for _ in params]
        self.vmax = [0 for _ in params]

        self.t = 0

    def step(self):
        self.t += 1
        for i, param in enumerate(self.params):
            if param.grad is None:
                continue

            grad = param.grad

            if self.maximize:
                grad = -grad

            param.data -= self.lr * self.weight_decay * param.data

            self.m[i] = self.B1 * self.m[i] + (1 - self.B1) * grad
            self.v[i] = self.B2 * self.v[i] + (1 - self.B2) * (grad**2)

            m_hat = self.m[i] / (1 - (self.B1**self.t))
            v_hat = self.v[i] / (1 - (self.B2**self.t))

            if self.amsgrad:
                self.vmax[i] = max(self.vmax[i], v_hat)
                param.data -= self.lr * m_hat / (self.vmax[i] ** 0.5 + self.eps)
            else:
                param.data -= self.lr * m_hat / (v_hat**0.5 + self.eps)



from mikegrad import Value
import mikegrad as mg
from mikegrad.nn import BaseModule
from typing import List


class BaseLoss(BaseModule):
    def __init__(self):
        pass


class MSELoss(BaseLoss):
    def __init__(self):
        super().__init__()

    def forward(self, y_pred, y_true):
        if len(y_pred) != len(y_true):
            raise ValueError("y_pred and y_true must have the same length")

        loss = sum([(y_pred[i] - y_true[i]) ** 2 for i in range(len(y_pred))]) / len(
            y_pred
        )
        return loss


class CrossEntropyLoss(BaseLoss):
    def __init__(self):
        super().__init__()

    def forward(
        self, y_pred: List[List[Value]], y_true: List[List[Value]] | List[Value]
    ):
        if len(y_pred) != len(y_true):
            raise ValueError("y_pred and y_true must have the same size")
        if type(y_true[0]) != list:
            y_true = mg.one_hot_encode(y_true, len(y_pred[0]))

        loss = sum(
            [
                -sum(
                    [y_true[i][j] * mg.log(y_pred[i][j]) for j in range(len(y_pred[i]))]
                )
                for i in range(len(y_pred))
            ]
        ) / len(y_pred)

        return loss


class BCELoss(BaseLoss):
    def __init__(self):
        super().__init__()

    def forward(self, y_pred: List[Value], y_true: List[Value]):
        if len(y_pred) != len(y_true):
            raise ValueError("y_pred and y_true must have the same length")

        loss = -sum(
            [
                y_true[i] * mg.log(y_pred[i]) + (1 - y_true[i]) * mg.log(1 - y_pred[i])
                for i in range(len(y_pred))
            ]
        ) / len(y_pred)

        return loss



from setuptools import setup, find_packages

setup(
    name="mikegrad",
    version="0.2.0",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.19.0",
    ],
    author="Can Michael Hucko",
    author_email="c.michaelhucko@gmail.com",
    description="A minimal autograd framework",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/canmike/mikegrad",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # Adjust based on your license
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",  # Adjust based on your minimum Python version
)



from .engine import Value
from .utils import (
    log,
    sigmoid,
    tanh,
    relu,
    leaky_relu,
    exp,
    one_hot_encode,
    from_list,
    to_list,
    from_numpy,
    to_numpy,
    from_pandas,
    argmax,
)

__all__ = [
    "Value",
    "nn",
    "losses",
    "log",
    "sigmoid",
    "tanh",
    "relu",
    "leaky_relu",
    "exp",
    "one_hot_encode",
    "from_list",
    "to_list",
    "from_numpy",
    "to_numpy",
    "from_pandas",
    "argmax",
]
